#if defined( DEBUG ) || defined( DEBUG_stdout )
#include <stdio.h>
#include <string.h>
#endif

#if defined( sgi )
#include <alloca.h>
#else
#include <stdlib.h>
#endif

#include "mpi.h"


/*@

  RGX_Swap_Task_Bndry_Strides - Swap/Exchange the data points in the 
                                ghost-point region of the local task grid
                                with data points in the core physical 
                                region of the task grids' nearest neighbors.

  Unmodified Input Variables :
. Comm - Input Communicator of all the tasks involved.
. Ndim - Dimensionality of the Grid (Integer).
. task_Grid_endpts - Integer Array, of size 2 x Ndim, of the two end-points'
                     coordinates of the LOCAL task grid.  It is implicitly
                     assumed to be declared as task_Grid_endpts[ 2 ][ Ndim ].
. stride_width - Integer Array, of size Ndim, of the stride width of the task
                 grid.  It is implicitly assumed to be declared as 
                 stride_width[ Ndim ].
. task_bndry_strides - MPI_Datatype Array, of size Ndim, represents the 
                       ghost-points region of the task grid.  It is implicitly 
                       assumed to be declared like task_bndry_strides[ Ndim ].
. Ngbrs - Integer Array, of size 2 x Ndim, of the task ID of the local task
          grid's nearest neighbors.  It is implicitly assumed to be declared 
          like Ngbrs[ Ndim ][ 2 ].
. MD_type - the MPI_Datatype of the data stored in the task grids, Task_Grid.

  Modified Output Variables :
. Task_Grid - Void pointer to the Block data array stored in the task grids. 
. returned value - returns MPI_SUCCESS if completed successfully
.n

.N RGX_fortran
.N RGX_minimal_set_of_variables
.N RGX_supported_datatype
.N RGX_yes_included

@*/

int RGX_Swap_Task_Bndry_Strides(
                                 MPI_Comm       Comm,
                                 int            Ndim,
                                 int           *task_Grid_endpts, 
                                 int           *stride_width,
                                 int           *task_bndry_strides, 
                                 int           *task_Ngbrs,
                                 MPI_Datatype   MD_type,

                                 void          *Task_Grid 
                               )
{

/*  Local Variables  */
MPI_Status  status;
int        *task_grid_sz;
int        *bndry_lgths;
int         lnr_adr_send;
int         lnr_adr_recv;
int        *Nplet_send;
int        *Nplet_recv;
void       *Task_Grid_send_ptr;
void       *Task_Grid_recv_ptr;
int         ary_size;
int         left_ngbr_idx, right_ngbr_idx;
int         MD_type_sz;
int         tagID;
int         ierr = MPI_ERR_UNKNOWN;
int         idir, ii;


#if defined( DEBUG ) || defined( DEBUG_stdout )
#include "debug.c.h"
#endif



/*  Allocate Memory for local variables  */

   ary_size     = Ndim * sizeof( int );
   task_grid_sz = ( int* ) alloca( ary_size );
   bndry_lgths  = ( int* ) alloca( ary_size );
   Nplet_send   = ( int* ) alloca( ary_size );
   Nplet_recv   = ( int* ) alloca( ary_size );

/*  Define the boundary lengths of the task grid  */

   for ( idir = 0; idir < Ndim; idir++ )
   {
      task_grid_sz[ idir ] = task_Grid_endpts[ idir + Ndim ]
                           - task_Grid_endpts[ idir ] + 1;
      bndry_lgths[ idir ]  = task_grid_sz[ idir ]
                           + 2 * stride_width[ idir ];
   }              /*  endif of { for ( idir = 0; idir < Ndim; idir++ ) }  */


#  if defined( DEBUG ) || defined( DEBUG_stdout )
      ierr_debug = MPI_Comm_rank( MPI_COMM_WORLD, &my_tid_debug );
#  endif

#  ifdef DEBUG
      sprintf( filename_debug, filename_debug_fmt, my_tid_debug );
      file_debug = fopen( filename_debug, "a+" );
#  endif

   MPI_Type_size( MD_type, &MD_type_sz );

   for ( idir = 0; idir < Ndim; idir++ )
   {
      if ( stride_width[ idir ] > 0 )
      {
         for ( ii = 0; ii < Ndim; ii++ )
         {
            if ( ii == idir )
            {
               Nplet_send[ ii ] = stride_width[ ii ];
               Nplet_recv[ ii ] = task_grid_sz[ ii ]
                                + stride_width[ ii ];
            }
            else
            {
               Nplet_send[ ii ] = 0;
               Nplet_recv[ ii ] = 0;
            }
         }

       /*  Define indexes to the left and right neigbor Task IDs  */
         left_ngbr_idx   = 2 * idir;
         right_ngbr_idx  = 2 * idir + 1;

       /*  It assumes use of C version adr_Nplet_2_lnr.c & adr_lnr_2_Nplet.c  */
       /*  Nplet[ii] in arg. list assumes starting from 0      */

         adr_Nplet_2_lnr( Ndim, bndry_lgths, Nplet_send, &lnr_adr_send );
         adr_Nplet_2_lnr( Ndim, bndry_lgths, Nplet_recv, &lnr_adr_recv );

#        ifdef DEBUG_stdout
            printf( "Task %d's 1st Nplet_send[] = [ ", my_tid_debug );
            for ( ii_dbg = 0; ii_dbg < Ndim; ii_dbg++ )
               printf( "%d ", Nplet_send[ii_dbg] );
            printf( "];  lnr_adr_send = %d,  idir = %d\n",
                    lnr_adr_send, idir );
            fflush( stdout );

            printf( "Task %d's 1st Nplet_recv[] = [ ", my_tid_debug );
            for ( ii_dbg = 0; ii_dbg < Ndim; ii_dbg++ )
               printf( "%d ", Nplet_recv[ii_dbg] );
            printf( "];  lnr_adr_recv = %d,  idir = %d\n",
                    lnr_adr_recv, idir );
            fflush( stdout );
#        endif

         Task_Grid_send_ptr = ( void* )( (char *)Task_Grid
                                       + lnr_adr_send * MD_type_sz );
         Task_Grid_recv_ptr = ( void* )( (char *)Task_Grid
                                       + lnr_adr_recv * MD_type_sz );
         tagID = 1;
         ierr = MPI_Sendrecv( Task_Grid_send_ptr, 1,
                              task_bndry_strides[ idir ],
                              task_Ngbrs[ left_ngbr_idx ], tagID,
                              Task_Grid_recv_ptr, 1,
                              task_bndry_strides[ idir ],
                              task_Ngbrs[ right_ngbr_idx ], tagID,
                              Comm, &status );
         if ( ierr ) return ierr;


         for ( ii = 0; ii < Ndim; ii++ )
         {
            if ( ii == idir )
            {
               Nplet_send[ ii ] = task_grid_sz[ ii ];
               Nplet_recv[ ii ] = 0;
            }
            else
            {
               Nplet_send[ ii ] = 0;
               Nplet_recv[ ii ] = 0;
            }
         }
         adr_Nplet_2_lnr( Ndim, bndry_lgths, Nplet_send, &lnr_adr_send ); 
         adr_Nplet_2_lnr( Ndim, bndry_lgths, Nplet_recv, &lnr_adr_recv ); 

#        ifdef DEBUG_stdout
            printf( "Task %d's 2nd Nplet_send[] = [ ", my_tid_debug ); 
            for ( ii_dbg = 0; ii_dbg < Ndim; ii_dbg++ )
               printf( "%d ", Nplet_send[ii_dbg] ); 
            printf( "];  lnr_adr_send = %d,  idir = %d\n",
                    lnr_adr_send, idir ); 
            fflush( stdout );

            printf( "Task %d's 2nd Nplet_recv[] = [ ", my_tid_debug ); 
            for ( ii_dbg = 0; ii_dbg < Ndim; ii_dbg++ )
               printf( "%d ", Nplet_recv[ii_dbg] ); 
            printf( "];  lnr_adr_recv = %d,  idir = %d\n",
                    lnr_adr_recv, idir ); 
            fflush( stdout );
#        endif

         Task_Grid_send_ptr = ( void* )( (char *)Task_Grid
                                       + lnr_adr_send * MD_type_sz );
         Task_Grid_recv_ptr = ( void* )( (char *)Task_Grid
                                       + lnr_adr_recv * MD_type_sz );
         tagID = 2;
         ierr = MPI_Sendrecv( Task_Grid_send_ptr, 1,
                              task_bndry_strides[ idir ],
                              task_Ngbrs[ right_ngbr_idx ], tagID,
                              Task_Grid_recv_ptr, 1,
                              task_bndry_strides[ idir ],
                              task_Ngbrs[ left_ngbr_idx ], tagID,
                              Comm, &status );
         if ( ierr ) return ierr;

      }           /*  endif of { if ( stride_width[ idir ] > 0 ) }  */

#     ifdef DEBUG
         fprintf( file_debug, "Process %d's array_Task_Grid[] :\n",
                  my_tid_debug );
         print_Task_Grid( file_debug, Ndim, bndry_lgths, MD_type, Task_Grid );
         fprintf( file_debug, "\n" );
         fflush( file_debug );
#     endif
      
   }              /*  for ( idir = 0; idir < Ndim; idir++ )  */

#  ifdef DEBUG
      fprintf( file_debug, "\n" );
      fclose( file_debug );
#  endif

   return MPI_SUCCESS;

}



#ifdef FortranNoUnderscore
#   define rgx_swap_task_bndry_strides__  rgx_swap_task_bndry_strides
#endif

#ifdef FortranSingleUnderscore
#   define rgx_swap_task_bndry_strides__  rgx_swap_task_bndry_strides_
#endif

/*  Fortran Interface for RGX_Swap_Task_Bndry_Strides()  */
void rgx_swap_task_bndry_strides__(
                                    int           *Comm,
                                    int           *Ndim,
                                    int           *task_Grid_endpts, 
                                    int           *stride_width,
                                    int           *task_bndry_strides, 
                                    int           *task_Ngbrs,
                                    int           *MD_type,

                                    void          *Task_Grid,
                                    int           *ierr
                                  )
{
   *ierr = RGX_Swap_Task_Bndry_Strides( *Comm, *Ndim, task_Grid_endpts,
                                        stride_width, task_bndry_strides,
                                        task_Ngbrs, *MD_type, Task_Grid );
}
